# Copyright (C) 2013 Deutsches Zentrum fuer Luft- und Raumfahrt(DLR, German Aerospace Center) <www.dlr.de>
"""
Created on 07.05.2014

@author: freu_se
"""


class DelisMMError(Exception):
    """classdocs"""
